﻿#include "model.h"
static int HaveAGoodDay;
namespace mm {
#line 1 "code/ChildMortality.mpp"
//LABEL(ChildMortality, EN) Child Mortality

/*  NOTE(ChildMortality, EN)

Added at Step 16

This module implements infant and child mortality (at ages 0-4).
Child mortality module is based on a proportional model consisting of a

- mortality baseline by age and sex
- relative risks by age for mothers characteristics: age at birth (3 groups), primary education (3 groups)
- a time trend by age 

The module is optional, and the user can choose between various options:

- Disable the child mortality model: the same model as for all ages is used 
- Child mortality model without alignment: the child mortality module replaces the overall mortality 
  module for ages 0-4. Note, that the overall life expectancy might be altered by this choice
- Child mortality model calibrated for an initial year, then trends as for other ages: In this choice,
  life expectancy is the same as in the overall mortality for the given year, but as the composition
  of the population by mothers' characteristics changes over time, the number of deaths (and therefore
  life expectancy) will be different for the following years, allowing for scenario comparisons.
- Child mortality model calibrated for an initial year, then specific trends from the child mortality 
  module: In this choice, life expectancy is the same as in the overall mortality for the given year, 
  but as the composition of the population by mothers' characteristics changes over time, and as a 
  result from potentially different trends, the number of deaths (and therefore life expectancy) 
  will be different for the following years, allowing for scenario comparisons.

If activated, the model for child mortality starts replacing the overall mortality model five
calendar years after the starting year of the simulation, ensuring that all children 'know' their 
mother's characteristics (are born in the simulation). As mother's characteristics are not available
for children born outside the country, immigrants are excluded from the model and handled by the
general mortality module. 
*/



//actor_set ChildMortalityCalibrator asTheChildMortalityCalibrator;
//link Person.lChildMortalityCalibrator ChildMortalityCalibrator.mlBabies[];

//STEP18
//link Person.lMother Person.mlChild[];			//EN Link Mother Children

//classification SELECTED_CHILD_MORTALITY_MODEL   //EN Child Mortality Model Options
//{
//    SCMM_MACRO,                                 //EN Disable child mortality model 
//    SCMM_MICRO_NOT_ALIGNED,                     //EN Child mortality model without alignment
//    SCMM_MICRO_ALIGNED_MACRO_TRENDS,            //EN Aligned, trends as for other ages
//    SCMM_MICRO_ALIGNED_MICRO_TRENDS             //EN Aligned, trends from child mortality model
//};

//classification CHILD_MORTALITY_RISKS            //EN Relative risks for child mortality
//{
//    CMR_AGE14,                                  //EN Mothers age at birth below 15        
//    CMR_AGE16,                                  //EN Mothers age at birth below 17
//    CMR_NOPRIM,                                 //EN Mother never entered primary education
//    CMR_PRIMDROP                                //EN Mother dropped out of primary education
//};

//classification MOTHER_AGE_CLASS                 //EN Mother age group
//{
//    MAC_14,                                     //EN 14 and below
//    MAC_16,                                     //EN 15 or 16
//    MAC_17P                                     //EN 17 and above
//};

//range CHILD_MORTALITY_AGE { 0, 4 };             //EN Age range for child mortality
//range CHILD_MORTALITY_YEARS{ 2018, 2113 };     //EN Year range for which child mortality is modeled

//parameters
//{
//    //EN Child mortality model selection
//    SELECTED_CHILD_MORTALITY_MODEL  SelectedChildMortalityModel;    
//    
//    //EN Child mortality baseline risk
//    double  ChildMortalityBaseRisk[CHILD_MORTALITY_AGE][SEX];
//
//    //EN Child mortality relative risk
//    double  ChildMortalityRelativeRisks[CHILD_MORTALITY_AGE][CHILD_MORTALITY_RISKS];
//
//    //EN Child mortality time trend
//    double ChildMortalityTrend[CHILD_MORTALITY_AGE][CHILD_MORTALITY_YEARS];
//};

//parameter_group PG_ChildMortality               //EN Child mortality
//{
//    SelectedChildMortalityModel, ChildMortalityBaseRisk,
//    ChildMortalityRelativeRisks, ChildMortalityTrend
//};

//actor ChildMortalityCalibrator
//{
//    double mort_male_0;                         //EN Child mortality risk male age 0
//    double mort_male_1;                         //EN Child mortality risk male age 1
//    double mort_male_2;                         //EN Child mortality risk male age 2
//    double mort_male_3;                         //EN Child mortality risk male age 3
//    double mort_male_4;                         //EN Child mortality risk male age 4
//
//    double mort_female_0;                       //EN Child mortality risk female age 0
//    double mort_female_1;                       //EN Child mortality risk female age 1
//    double mort_female_2;                       //EN Child mortality risk female age 2
//    double mort_female_3;                       //EN Child mortality risk female age 3
//    double mort_female_4;                       //EN Child mortality risk female age 4
//
//    logical calibration_is_done = { FALSE };    //EN Calibration is done
//    void Start();                               //EN Starts the calibrator actor
//
//    //EN Child mortality calibration event
//    event timeChildMortalityCalibrationEvent, ChildMortalityCalibrationEvent;    
//
//    integer int_age = self_scheduling_int(age);
//};

void ChildMortalityCalibrator::Start()
{ if (__finished) {StartAfterFinish(); return; }; StartSpecial1();
     poStateFunctions->Set_time(  MIN(SIM_YEAR_RANGE), true );
StartSpecial2();}

TIME ChildMortalityCalibrator::timeChildMortalityCalibrationEvent()
{
    // if not done yet and a model with alignment is chosen
    if (! 
        calibration_is_done && ( gprParam->SelectedChildMortalityModel == SCMM_MICRO_ALIGNED_MACRO_TRENDS 
        || gprParam->SelectedChildMortalityModel == SCMM_MICRO_ALIGNED_MICRO_TRENDS))
    // return the moment in which the calibration is to be performed
    {
        return  MIN(CHILD_MORTALITY_YEARS);
    }
    // else never call the calibration event
    else return TIME_INFINITE;
}

void ChildMortalityCalibrator::ChildMortalityCalibrationEvent()
{
    // local matrices
    double dDeaths[SIZE(CHILD_MORTALITY_AGE)][SIZE(SEX)];      // Number expected deaths
    double dProb[SIZE(CHILD_MORTALITY_AGE)][SIZE(SEX)];        // Death probability
    double dBase[SIZE(CHILD_MORTALITY_AGE)][SIZE(SEX)];        // Baseline Hazard found in simulation
    double dRelRisks[SIZE(CHILD_MORTALITY_AGE)][SIZE(PRIMARY_LEVEL)][SIZE(MOTHER_AGE_CLASS)];    // Relative risks
    long nPop[SIZE(CHILD_MORTALITY_AGE)][SIZE(SEX)][SIZE(PRIMARY_LEVEL)][SIZE(MOTHER_AGE_CLASS)];// Pop sizes

    // Initialize matrices
    // set population sizes nPop and expected deaths nDeaths to 0
    // sets death probabilities dProb by age and sex for the year in which the model is calibrated
    for (int nAge = 0; nAge < SIZE(CHILD_MORTALITY_AGE); nAge++)
    {
        for (int nSex = 0; nSex < SIZE(SEX); nSex++)
        {
            dProb[nAge][nSex] = 1.0 - exp(-gprParam->MortalityTable[CHKBND(nAge,101,"MortalityTable",0)][CHKBND(nSex,2,"MortalityTable",1)] 
                * gprParam->MortalityTrend[CHKBND(RANGE_POS(SIM_YEAR_RANGE, MIN(CHILD_MORTALITY_YEARS)),101,"MortalityTrend",0)][CHKBND(nSex,2,"MortalityTrend",1)]);
            dDeaths[nAge][nSex] = 0.0;
            for (int nEduc = 0; nEduc < SIZE(PRIMARY_LEVEL); nEduc++)
            {
                for (int nMotherAge = 0; nMotherAge < SIZE(MOTHER_AGE_CLASS); nMotherAge++)
                {
                    nPop[nAge][nSex][nEduc][nMotherAge] = 0.0;
                }
            }
        }
    }

    // Initialize relative risks dRelRisks by age, mothers education, and mothers age
    for (int nAge = 0; nAge < SIZE(CHILD_MORTALITY_AGE); nAge++)    
    {
        dRelRisks[nAge][PL_NO][MAC_14]      = gprParam->ChildMortalityRelativeRisks[CHKBND(nAge,5,"ChildMortalityRelativeRisks",0)][CHKBND(CMR_NOPRIM,4,"ChildMortalityRelativeRisks",1)] 
            * gprParam->ChildMortalityRelativeRisks[CHKBND(nAge,5,"ChildMortalityRelativeRisks",0)][CHKBND(CMR_AGE14,4,"ChildMortalityRelativeRisks",1)];
        dRelRisks[nAge][PL_ENTER][MAC_14]   = gprParam->ChildMortalityRelativeRisks[CHKBND(nAge,5,"ChildMortalityRelativeRisks",0)][CHKBND(CMR_PRIMDROP,4,"ChildMortalityRelativeRisks",1)] 
            * gprParam->ChildMortalityRelativeRisks[CHKBND(nAge,5,"ChildMortalityRelativeRisks",0)][CHKBND(CMR_AGE14,4,"ChildMortalityRelativeRisks",1)];
        dRelRisks[nAge][PL_GRAD][MAC_14]    = gprParam->ChildMortalityRelativeRisks[CHKBND(nAge,5,"ChildMortalityRelativeRisks",0)][CHKBND(CMR_AGE14,4,"ChildMortalityRelativeRisks",1)];
        dRelRisks[nAge][PL_NO][MAC_16]      = gprParam->ChildMortalityRelativeRisks[CHKBND(nAge,5,"ChildMortalityRelativeRisks",0)][CHKBND(CMR_NOPRIM,4,"ChildMortalityRelativeRisks",1)] 
            * gprParam->ChildMortalityRelativeRisks[CHKBND(nAge,5,"ChildMortalityRelativeRisks",0)][CHKBND(CMR_AGE16,4,"ChildMortalityRelativeRisks",1)];
        dRelRisks[nAge][PL_ENTER][MAC_16]   = gprParam->ChildMortalityRelativeRisks[CHKBND(nAge,5,"ChildMortalityRelativeRisks",0)][CHKBND(CMR_PRIMDROP,4,"ChildMortalityRelativeRisks",1)] 
            * gprParam->ChildMortalityRelativeRisks[CHKBND(nAge,5,"ChildMortalityRelativeRisks",0)][CHKBND(CMR_AGE16,4,"ChildMortalityRelativeRisks",1)];
        dRelRisks[nAge][PL_GRAD][MAC_16]    = gprParam->ChildMortalityRelativeRisks[CHKBND(nAge,5,"ChildMortalityRelativeRisks",0)][CHKBND(CMR_AGE16,4,"ChildMortalityRelativeRisks",1)];
        dRelRisks[nAge][PL_NO][MAC_17P]     = gprParam->ChildMortalityRelativeRisks[CHKBND(nAge,5,"ChildMortalityRelativeRisks",0)][CHKBND(CMR_NOPRIM,4,"ChildMortalityRelativeRisks",1)];
        dRelRisks[nAge][PL_ENTER][MAC_17P]  = gprParam->ChildMortalityRelativeRisks[CHKBND(nAge,5,"ChildMortalityRelativeRisks",0)][CHKBND(CMR_PRIMDROP,4,"ChildMortalityRelativeRisks",1)];
        dRelRisks[nAge][PL_GRAD][MAC_17P]   = 1.0;
    }

    // Population Count
    // calculates population sizes nPop by age, sex, mothers education, and mothers age group
    // calculates expected deaths dDeaths by age and sex
    for (long nJ = 0; nJ < asAllPersons->Count(); nJ++)
    {
        Person * prPerson = new Person();
        prPerson = asAllPersons->Item(nJ);
        if (prPerson-> integer_age <= MAX(CHILD_MORTALITY_AGE) && prPerson->mother_known )
        {
            for (int nAge = 0; nAge < SIZE(CHILD_MORTALITY_AGE); nAge++)
            {
                for (int nSex = 0; nSex < SIZE(SEX); nSex++)
                {
                    if (nSex == prPerson-> sex && nAge == prPerson->integer_age )
                    {
                        dDeaths[nAge][nSex] = dDeaths[nAge][nSex] + dProb[nAge][nSex];
                        if (prPerson-> mother_educ_0 == TRUE)
                        {
                            if (prPerson-> mother_max_14 == TRUE) nPop[nAge][nSex][PL_NO][MAC_14]++;
                            else if (prPerson-> mother_max_16 == TRUE) nPop[nAge][nSex][PL_NO][MAC_16]++;
                            else  nPop[nAge][nSex][PL_NO][MAC_17P]++;
                        }
                        else if (prPerson-> mother_educ_1 == TRUE)
                        {
                            if (prPerson-> mother_max_14 == TRUE) nPop[nAge][nSex][PL_ENTER][MAC_14]++;
                            else if (prPerson-> mother_max_16 == TRUE) nPop[nAge][nSex][PL_ENTER][MAC_16]++;
                            else  nPop[nAge][nSex][PL_ENTER][MAC_17P]++;
                        }
                        else
                        {
                            if (prPerson-> mother_max_14 == TRUE) nPop[nAge][nSex][PL_GRAD][MAC_14]++;
                            else if (prPerson-> mother_max_16 == TRUE) nPop[nAge][nSex][PL_GRAD][MAC_16]++;
                            else  nPop[nAge][nSex][PL_GRAD][MAC_17P]++;
                        }
                    }
                }
            }
        }
    }

    // find calibrated baselines
    for (int nAge = 0; nAge < SIZE(CHILD_MORTALITY_AGE); nAge++)
    {
        for (int nSex = 0; nSex < SIZE(SEX); nSex++)
        {
            double  dUpper = 2.0;
            double  dLower = 0.0;
            double  dCenter = 1.0;
            double  dNumberDeaths = 0.0;
            int     nIterations = 10000;
            
            while ( abs(dNumberDeaths - dDeaths[nAge][nSex]) > 0.0001 && nIterations > 0 )
            {
                nIterations--;
                dBase[nAge][nSex] = ( dLower + dUpper) / 2.0;
            
                dNumberDeaths = 0.0;

                //Celculate numer of deaths for given dBase
                for (int nEduc = 0; nEduc < SIZE(PRIMARY_LEVEL); nEduc++)
                {
                    for (int nMotherAge = 0; nMotherAge < SIZE(MOTHER_AGE_CLASS); nMotherAge++)
                    {
                        dNumberDeaths = dNumberDeaths + nPop[nAge][nSex][nEduc][nMotherAge] 
                            * (1-exp(-dBase[nAge][nSex] * dRelRisks[nAge][nEduc][nMotherAge]));
                        double hugo = nPop[nAge][nSex][nEduc][nMotherAge];
                        double emil = dBase[nAge][nSex];
                        double gustav = dRelRisks[nAge][nEduc][nMotherAge];
                        double hannelore = dDeaths[nAge][nSex];
                        double anton = 0;

                    }
                }
                // shrink search interval
                if ( dNumberDeaths > dDeaths[nAge][nSex]) dUpper = dBase[nAge][nSex];
                else dLower = dBase[nAge][nSex];
            }
            int hugo = nIterations;
        }
    }

    // set states
     poStateFunctions->Set_mort_male_0(  dBase[0][MALE] );  poStateFunctions->Set_mort_female_0(  dBase[0][FEMALE] );
     poStateFunctions->Set_mort_male_1(  dBase[1][MALE] );  poStateFunctions->Set_mort_female_1(  dBase[1][FEMALE] );
     poStateFunctions->Set_mort_male_2(  dBase[2][MALE] );  poStateFunctions->Set_mort_female_2(  dBase[2][FEMALE] );
     poStateFunctions->Set_mort_male_3(  dBase[3][MALE] );  poStateFunctions->Set_mort_female_3(  dBase[3][FEMALE] );
     poStateFunctions->Set_mort_male_4(  dBase[4][MALE] );  poStateFunctions->Set_mort_female_4(  dBase[4][FEMALE] );

     poStateFunctions->Set_calibration_is_done(  TRUE );
}

//actor Person
//{
//    //EN Child mortality overrides general mortality module
//    logical child_mortality_overrides = calendar_year >= MIN(CHILD_MORTALITY_YEARS)
//        && SelectedChildMortalityModel != SCMM_MACRO
//        && integer_age <= MAX(CHILD_MORTALITY_AGE)
//        && mother_known;
//
//    logical mother_known = { FALSE };                   //EN Mother is known
//    logical mother_educ_0 = { FALSE };                  //EN Mother never entered primary school
//    logical mother_educ_1 = { FALSE };                  //EN Mother dropped out of primary school
//    logical mother_max_14 = { FALSE };                  //EN Mother was 14 or below at birth
//    logical mother_max_16 = { FALSE };                  //EN Mother was 15 or 16 at birth
//
//    PARITY_RANGE infant_deaths = { 0 };                 //EN Infant deaths experienced
//
//    //EN Baseline mortality
//    double child_mortality = (lChildMortalityCalibrator == NULL) ? 0.0 :
//        (integer_age == 0 && sex == MALE) ? lChildMortalityCalibrator->mort_male_0 :
//        (integer_age == 1 && sex == MALE) ? lChildMortalityCalibrator->mort_male_1 :
//        (integer_age == 2 && sex == MALE) ? lChildMortalityCalibrator->mort_male_2 :
//        (integer_age == 3 && sex == MALE) ? lChildMortalityCalibrator->mort_male_3 :
//        (integer_age == 4 && sex == MALE) ? lChildMortalityCalibrator->mort_male_4 :
//        (integer_age == 0 && sex == FEMALE) ? lChildMortalityCalibrator->mort_female_0 :
//        (integer_age == 1 && sex == FEMALE) ? lChildMortalityCalibrator->mort_female_1 :
//        (integer_age == 2 && sex == FEMALE) ? lChildMortalityCalibrator->mort_female_2 :
//        (integer_age == 3 && sex == FEMALE) ? lChildMortalityCalibrator->mort_female_3 :
//        (integer_age == 4 && sex == FEMALE) ? lChildMortalityCalibrator->mort_female_4 : 0.0;
//
//    //EN Child mortality event
//    event timeChildMortalityEvent, ChildMortalityEvent; 
//};

TIME Person::timeChildMortalityEvent()
{
    double dEventTime = TIME_INFINITE;
    double dHazard = 0.0;

    if (child_mortality_overrides )    // at risk (the model is used and overrides the overall mortality)
    {
        // base risk * trend
        // unaligned model
        if (gprParam->SelectedChildMortalityModel == SCMM_MICRO_NOT_ALIGNED) 
        {
            dHazard = gprParam->ChildMortalityBaseRisk[CHKBND(integer_age ,5,"ChildMortalityBaseRisk",0)][CHKBND(sex ,2,"ChildMortalityBaseRisk",1)]
                * gprParam->ChildMortalityTrend[CHKBND(integer_age ,5,"ChildMortalityTrend",0)][CHKBND(RANGE_POS(CHILD_MORTALITY_YEARS, calendar_year ),96,"ChildMortalityTrend",1)];
        }
        
        // aligned with overall trend
        else if (gprParam->SelectedChildMortalityModel == SCMM_MICRO_ALIGNED_MACRO_TRENDS) 
        {
            dHazard = 
                child_mortality * gprParam->MortalityTrend[CHKBND(RANGE_POS(SIM_YEAR_RANGE, calendar_year ),101,"MortalityTrend",0)][CHKBND(sex ,2,"MortalityTrend",1)]
                / gprParam->MortalityTrend[CHKBND(RANGE_POS(SIM_YEAR_RANGE, MIN(CHILD_MORTALITY_YEARS)),101,"MortalityTrend",0)][CHKBND(sex ,2,"MortalityTrend",1)];
        }

        // aligned with specific child mortality trend
        else 
        {
            dHazard = 
                child_mortality * gprParam->ChildMortalityTrend[CHKBND(integer_age ,5,"ChildMortalityTrend",0)][CHKBND(RANGE_POS(CHILD_MORTALITY_YEARS, calendar_year ),96,"ChildMortalityTrend",1)];
        }

        // relativ risks
        if (mother_educ_0 ) dHazard = dHazard * gprParam->ChildMortalityRelativeRisks[CHKBND(integer_age ,5,"ChildMortalityRelativeRisks",0)][CHKBND(CMR_NOPRIM,4,"ChildMortalityRelativeRisks",1)];
        else if (mother_educ_1 ) dHazard = dHazard * gprParam->ChildMortalityRelativeRisks[CHKBND(integer_age ,5,"ChildMortalityRelativeRisks",0)][CHKBND(CMR_PRIMDROP,4,"ChildMortalityRelativeRisks",1)];
        if (mother_max_14 ) dHazard = dHazard * gprParam->ChildMortalityRelativeRisks[CHKBND(integer_age ,5,"ChildMortalityRelativeRisks",0)][CHKBND(CMR_AGE14,4,"ChildMortalityRelativeRisks",1)];
        else if (mother_max_16 ) dHazard = dHazard * gprParam->ChildMortalityRelativeRisks[CHKBND(integer_age ,5,"ChildMortalityRelativeRisks",0)][CHKBND(CMR_AGE16,4,"ChildMortalityRelativeRisks",1)];

        // if hazard is positive calculate event time
        if (dHazard > 0) dEventTime = WAIT(-log(RandUniform(4)) / dHazard);;
    }
    return dEventTime;
}

void Person::ChildMortalityEvent()
{
    MortalityEvent();
}

} // namespace 
