﻿#include "model.h"
static int HaveAGoodDay;
namespace mm {
#line 1 "code/Mortality.mpp"
//LABEL(Mortality, EN) Mortality Module

/*	NOTE(Mortality, EN)

    Changed at step 5

	This module implements a simple model of mortality.
	People can die at any moment of time based on age-specific mortality rates.
    Age specific mortality rates are calculated from a standard life table and a trend factor.
    The trend factor is found in pre-simulation and calibrates mortality to reach a target
    period life expectancy. The standard life table and target life expectancies are model parameters.
	The only state of this module is 'alive' which is set to 'false' at death.
	Also at death, the Modgen function Finish() is called which clears up memory space.
*/

// Parameters

//parameters
//{
//    double                  MortalityTable[AGE_RANGE][SEX];	            //EN Mortality hazard by age
//    double					LifeExpectancy[SIM_YEAR_RANGE][SEX];		//EN Life Expectancy
//    model_generated double	MortalityTrend[SIM_YEAR_RANGE][SEX];		//EN Mortality trend
//};

//parameter_group PG02_OverallMortality                                   //EN Overall mortality
//{
//    MortalityTable, LifeExpectancy
//};

//parameter_group PG02_Mortality                                          //EN Mortality
//{
//    SelectedChildMortalityModel,
//    PG02_OverallMortality,
//    PG_ChildMortality
//};

// Actor definitions

//actor Person				
//{
//	// States
//
//	logical alive = { TRUE };						//EN Alive
//	
//	// Events
//
//	event	timeMortalityEvent, MortalityEvent;		//EN Mortality event
//};

// Implementation of functions and events

/*	NOTE(Person.MortalityEvent, EN)

    Changed at step 2

	This event implements an age-specific hazard of death. 
	The time function is very typical for time functions scheduling an event in continuouse time.
	If a person is at risk of the event, a random waiting time is drawn and returned. 
	Note that whenever a state that influences the waiting time changes (in this case only integer_age),
	a new waiting time is drawn automatically based on the updated rate and the event is automatically re-scheduled.

*/

TIME Person::timeMortalityEvent()
{
    TIME    dEventTime = TIME_INFINITE;
    double  dMortalityHazard
        = gprParam->MortalityTable[CHKBND(integer_age ,101,"MortalityTable",0)][CHKBND(sex ,2,"MortalityTable",1)] * gprParam->MortalityTrend[CHKBND(RANGE_POS(SIM_YEAR_RANGE, calendar_year ),101,"MortalityTrend",0)][CHKBND(sex ,2,"MortalityTrend",1)];

    // check if a person is at risk  
    if (dMortalityHazard > 0.0 &&  in_projected_time && !child_mortality_overrides )
    {
        // determine the event time
        // the formula [ -log(rand) / hazard ] calculates an exponentially distributed waiting time 
        // based on a uniform distributed random number and a given hazard rate
        dEventTime = WAIT(-log(RandUniform(1)) / dMortalityHazard);
    }
    // return the event time, if the maximum age is not reached at that point
    if (dEventTime <  time_of_birth + MAX(AGE_RANGE) + 1.0) return dEventTime;
    // otherwise, return the moment, at which the maximum age is reached
    else return  time_of_birth + MAX(AGE_RANGE) + 0.9999;
}

void Person::MortalityEvent()
{
	 poStateFunctions->Set_alive(  FALSE );
    //STEP18
    if ( lMother != NULL &&  age < 5.0)
    {
        lMother ->UpdateTime() -> poStateFunctions->PlusPlusSet_infant_deaths();
    }
	Finish();										// Remove the actor from the simulation.
}

void PreSimulation4()
{
    double	dLower, dUpper, dCenter, dResult, dTarget, dAlive, dDeaths;
    int		nIterations;
    for (int nSex = 0; nSex < SIZE(SEX); nSex++)
    {
        for (int nYear = 0; nYear < SIZE(SIM_YEAR_RANGE); nYear++)
        {
            dTarget = gprParam->LifeExpectancy[CHKBND(nYear,101,"LifeExpectancy",0)][CHKBND(nSex,2,"LifeExpectancy",1)];      // Target: life expectancy
            dResult = 0.0;                              // Search result: life expectancy  
            nIterations = 10000;                        // Maximum number of iterations in search
            dLower = 0.1;                               // Lower limit of calibration factor (relative risk)
            dUpper = 3.0;                               // Upper limit of calibration factor (relative risk)

            while (abs(dResult - dTarget) > 0.0001 && nIterations > 0)
            {
                nIterations--;
                dCenter = (dLower + dUpper) / 2.0;      // New calibration factor for probing

                dResult = 0.0;
                dAlive = 1.0;                           // Proportion of people still alive    

                // Life expectancy calculated applying calibration factor
                for (int nAge = 0; nAge < SIZE(AGE_RANGE); nAge++)
                {
                    // proportion of deaths in year: survival = exp(-hazard)
                    dDeaths = dAlive * (1 - exp(-gprParam->MortalityTable[CHKBND(nAge,101,"MortalityTable",0)][CHKBND(nSex,2,"MortalityTable",1)] * dCenter));
                    dAlive = dAlive - dDeaths;
                    // people dying in this year are assumed to die in th emiddle of the year
                    dResult = dResult + dDeaths * 0.5 + dAlive;
                }
                // Moving the search limits for next iteration 
                if (dTarget < dResult) dLower = dCenter;
                else dUpper = dCenter;
            }
            // copying the best solution into the model-generated mortality trend parameter
            gprParam->MortalityTrend[CHKBND(nYear,101,"MortalityTrend",0)][CHKBND(nSex,2,"MortalityTrend",1)] = dCenter;
        }
    }
}
} // namespace 
